// Copyright (C) 2005, 2007 International Business Machines and others.
// All Rights Reserved.
// This code is published under the Common Public License.
//
// $Id: hs071_main.cpp 1245 2008-05-30 03:46:59Z andreasw $
//
// Authors:  Carl Laird, Andreas Waechter     IBM    2005-08-10

#include "IpIpoptApplication.hpp"
#include "bl.hpp"

using namespace Ipopt;

int main (int argc, char *argv[])
{
    // Create a new instance of your nlp
    //  (use a SmartPtr, not raw)


    // my variables
    char struct_dir[1000];
    char initial_params_file[1000];
    int num_parameters;
    char newdir[1000];
    double thermo_precision;
    double regularizer_precision;
    int regularizer_mean;
    char model_options_file[1000];
    int num_constraints = 0;
    int num_jacobian = 0;
    int change_variables = 1;
    int use_local_machine = 1;  // if it's 0, send it to a cluster
    double bounds;
    char Tfile[1000];
    char thermoAfile[1000];
    char thermoBfile[1000];
    char Tconmat_file[1000];
    char cluster[100];
    int include_dangles = 1;
    int add_gaussian_regularizer = 0;

    if (argc != 19)
    {
        printf ("Usage: %s <use_local_machine:0|1> <num_params> <input_training_dir> <thermo_precision> <regularizer_mean:0|1> <regularizer_precision> <input_initial_params> <newdir> <model_options_file> <change_variables> <bounds> <Tfile> <thermoAfile> <thermoBfile> <Tconmat_file> <cluster> <include_dangles> <add_gaussian_regularizer>\n", argv[0]);
        printf ("<reg_precision> should be 0 if we don't want any regularizer\n");
        exit(1);
    }

    use_local_machine = atoi (argv[1]);
    num_parameters = atoi (argv[2]);
    strcpy (struct_dir, argv[3]);
    thermo_precision = atof (argv[4]);
    regularizer_mean = atoi (argv[5]);
    regularizer_precision = atof (argv[6]);
    strcpy (initial_params_file, argv[7]);
    strcpy (newdir, argv[8]);
    strcpy (model_options_file, argv[9]);
    change_variables = atoi (argv[10]);
    bounds = atof (argv[11]);
    strcpy (Tfile, argv[12]);
    strcpy (thermoAfile, argv[13]);
    strcpy (thermoBfile, argv[14]);
    strcpy (Tconmat_file, argv[15]);
    strcpy (cluster, argv[16]);
    include_dangles = atoi (argv[17]);
    add_gaussian_regularizer = atoi (argv[18]);

    SmartPtr<TNLP> mynlp;
    
    if (!include_dangles)
    {    
        num_constraints = 0;
        num_jacobian = 0;
    }
    else        //if (num_parameters == 363)
    {
        if (change_variables)
        {
            num_constraints = 192;
            num_jacobian = num_constraints*num_parameters;     // it's essentially a dense matrix
        } 
        else
        {
            num_constraints = 144;    // there are 144 constraints of type x - y <= 0
            // Also, there are 48 bounds of type x <= 0
            num_jacobian = 144*2;
        }
    }

    mynlp = new BL(use_local_machine, num_parameters, num_constraints, num_jacobian, struct_dir, thermo_precision, regularizer_mean, regularizer_precision, initial_params_file, newdir, model_options_file, change_variables, bounds, Tfile, thermoAfile, thermoBfile, Tconmat_file, cluster, add_gaussian_regularizer);
    
    
    // Create a new instance of IpoptApplication
    //  (use a SmartPtr, not raw)
    SmartPtr<IpoptApplication> app = new IpoptApplication();
    
    // Change some options
    // Note: The following choices are only examples, they might not be
    //       suitable for your optimization problem.
    app->Options()->SetNumericValue("tol", 1e-4);
    app->Options()->SetStringValue("mu_strategy", "adaptive");
    app->Options()->SetStringValue("hessian_approximation", "limited-memory");
    
    //app->Options()->SetStringValue("output_file", "ipopt.out");
    // The following overwrites the default name (ipopt.opt) of the
    // options file
    //app->Options()->SetStringValue("option_file_name", "ipopt.opt");
    
    // Intialize the IpoptApplication and process the options
    ApplicationReturnStatus status;
    status = app->Initialize();
    if (status != Solve_Succeeded) {
        printf("\n\n*** Error during initialization!\n");
        return (int) status;
    }
    
    // Ask Ipopt to solve the problem
    status = app->OptimizeTNLP(mynlp);
    
    if (status == Solve_Succeeded) {
        printf("\n\n*** The problem solved!\n");
    }
    else {
        printf("\n\n*** The problem FAILED!\n");
    }
    
    // As the SmartPtrs go out of scope, the reference count
    // will be decremented and the objects will automatically
    // be deleted.
    
    return (int) status;
}
