#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "constants.h"
#include "simfold.h"
#include "common.h"
#include "externs.h"
#include "params.h"

//#define MAXSLEN 4000
#define ABERANT 99999

void my_strcpy (char *source, char *dest, int from, int len)
{
  int i;
  for (i=0; i < len; i++)
    {
      dest[i] = source[from+i];
    }
  dest[i] = '\0';
}

void my_strcat (char *source, char *dest, int from, int len)
{
  int i, l;
  l = strlen (dest);  
  for (i=0; i < len; i++)
    {
      dest[l+i] = source[from+i];
    }
  dest[l+i] = '\0';
}


int main(int argc, char *argv[])
{
    if (argc != 7)
    {
        printf ("Usage: %s <xml_filename> <model_options_file> <initial_parameter_file> <use_published_errors:0|1> <second_order_filename> <first_order_filename>\n", argv[0]);
        printf ("The first order matrix will also contain the constant at the end\n");
        exit(0);
    }
    
            
    char filename[1000];
    char options_model_file[1000];  // "none" if the Turner99-363 model is used, AND it's compiled without -DEXTENDEDMODEL
    char params_file[1000];
    int i, j;
    char sequence[MAXSLEN];
    char structure[MAXSLEN];
    
    // sometimes the initial state is not the completely unfolded sequence. In this case, sequence0 and structure0 are the initial state
    char sequence0[MAXSLEN];
    char structure0[MAXSLEN];    
    
    int self_comp;
    double lb, ub, dg_37, std_dg_37=ABERANT, err_dg_37=ABERANT;
    char buffer [10000];
    // configuration file
    char config_file[200] = "../../MultiRNAFold/params/multirnafold.conf";
    double energy;
    double counter_min[MAXNUMPARAMS];
    int param_met[MAXNUMPARAMS];
    char experiment_for_param[MAXNUMPARAMS][100];
    double f_min;
    double f0;    

    double counter0[MAXNUMPARAMS];    // just in case we have sequence0 and structure0
    char solver[10];
    int res;
    FILE *xml; 
    int constraints_only;
    char objective[20];
    char exp_id[100];
    double precision;
    char second_order_filename[1000];
    char first_order_filename[1000];
    int use_published_errors;
    
    simple_internal_energy = 0;
    strcpy (filename, argv[1]);
    strcpy (options_model_file, argv[2]);
    strcpy (params_file, argv[3]);
    use_published_errors = atoi (argv[4]);
    strcpy (second_order_filename, argv[5]);
    strcpy (first_order_filename, argv[6]);
    constraints_only = 1;
    strcpy (objective, "l2norm");
    strcpy (solver, "CPLEX");
    simple_internal_energy = 0;
    char constraint_comment[5000];

    if (strcmp (options_model_file, "none") != 0)
        read_parsi_options_from_file (options_model_file);
    
    // initialize the thermodynamic parameters
    // call init_data only once for the same dna_or_rna and same temperature
    // if one of them changes, call init_data again (see below)
    init_data (argv[0], config_file, RNA, 37.0);   
       
    num_params = create_string_params();
    simple_dangling_ends = 1;
    
    //  for (i=0; i < num_params; i++)
    //     printf ("%d\t%s\n", i+1, string_params[i]);
    
    //  exit(0);

    // I have to call the following function
    fill_data_structures_with_new_parameters (params_file);

    // I need this, otherwise I don't know why params are "experimental additions" and which not -- could be taken from a file, of course
    num_params = create_string_params();
    set_starters();
    creating_model = 1;
    fill_similarity_rule_with_optical_melting_reference (filename);
    creating_model = 0;
    fill_data_structures_with_new_parameters (params_file);
    num_params = create_string_params();
    /////////////
    

    // Sometimes get the second order matrix 
    int compute_second_order_matrix = 1;
    double **second_order_matrix;
    double *first_order;
    double zero_order;
    if (compute_second_order_matrix)
    {
        second_order_matrix = new double *[num_params];
        for (i=0; i < num_params; i++)
        {
            second_order_matrix[i] = new double[num_params];
            for (j=0; j < num_params; j++)
            {
                second_order_matrix[i][j] = 0;
            }
        }
        first_order = new double[num_params];
        for (i=0; i < num_params; i++)  first_order[i] = 0;
        zero_order = 0;
    }
            
    for (i=0; i < num_params; i++)    
    {
        counter0[i] = 0;
        param_met[i] = 0;
        experiment_for_param[i][0] = '\0';
    }   

    
    if ((xml = fopen (filename, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", filename);
        exit (0);
    }           
    
    int k;
    k = 0;
    int delta_index = 1;
    double margin;
    int counter = 0;
    int num_experiments;    // = 686;    //    506;    //524;
    
    num_experiments = 0;
    // get num_experiments
    while (!feof (xml))
    {    
        // the file is in XML format, among whose lines it has the following types:
        //    <EXPERIMENT sequence="CCGG CCGG" structure="(((( ))))" dG-37="-4.36" STD-dG-37="0.1"> </EXPERIMENT>
        //    <EXPERIMENT sequence="UGACCUCA UGAGGUCA" structure="(((((((( ))))))))" dG-37="-12.34" ERR-dG-37="3%"> </EXPERIMENT>
        char *begin;
        fgets (buffer, sizeof(buffer), xml);
        begin = strstr (buffer, "<EXPERIMENT");        
        if (begin == NULL) continue;    
        num_experiments++;
    }
    fclose (xml);
    
    //printf ("Num exp: %d\n", num_experiments);
    
    if ((xml = fopen (filename, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", filename);
        exit (0);
    }               

    while (!feof (xml))
    {    
        // the file is in XML format, among whose lines it has the following types:
        //    <EXPERIMENT sequence="CCGG CCGG" structure="(((( ))))" dG-37="-4.36" STD-dG-37="0.1"> </EXPERIMENT>
        //    <EXPERIMENT sequence="UGACCUCA UGAGGUCA" structure="(((((((( ))))))))" dG-37="-12.34" ERR-dG-37="3%"> </EXPERIMENT>
        
        // try including only the first paper. With everything I can't save the solution because it says "No basic solution"
        //    Problem solved: In the cplex script, I have to write: "write file.vec"
        //if (counter >= 94) break;
        
        fgets (buffer, sizeof(buffer), xml);
        //printf ("%s", buffer);                         
        char *begin;
        char temp[100];
        sequence[0] = '\0'; structure[0] = '\0'; dg_37 = ABERANT; err_dg_37 = ABERANT; std_dg_37 = ABERANT;
        begin = strstr (buffer, "<EXPERIMENT");        
        if (begin == NULL) continue;
        
        sequence0[0] = '\0';
        structure0[0] = '\0';
        get_data_from_buffer (buffer, "id=\"", '\"', exp_id);
        get_data_from_buffer (buffer, "sequence0=\"", '\"', sequence0);        
        get_data_from_buffer (buffer, "structure0=\"", '\"', structure0);        
        get_data_from_buffer (buffer, "sequence=\"", '\"', sequence);
        get_data_from_buffer (buffer, "structure=\"", '\"', structure);
        get_data_from_buffer (buffer, "dG-37=\"", '\"', temp);  dg_37 = atof (temp);
        if (get_data_from_buffer (buffer, "STD-dG-37=\"", '\"', temp))  std_dg_37 = atof (temp);
        if (get_data_from_buffer (buffer, "ERR-dG-37=\"", '%', temp))  err_dg_37 = atof (temp);                                      
           
        //strcpy (sequence,  "GCCGGA GCCGGA");
        //strcpy (structure, ".((((. .)))).");
        //printf ("%s\t%s\n", sequence, structure);
        
        // Self-complementarity doesn't have to do anything with dG. Symmetry does! 
        // If symmetric, the 0.43 penalty is part of the free energy value.
        // Not exactly sure what symmetric means.
        //  Following conversations with Dave Mathews,
        //      I consider that 2 sequences are symmetric if they are perfectly complementary and identical
        
        
        // first, adjust dg_37 according to self-complementarity
        //  Let S1 be the first sequence, let S2 be the second, and let S3 be the reverse complement of S1. 
        //    I considered the S1/S2 duplex is self-complementary if S1, S2 and S3 are identical.
        // This is incorect!! The 0.43 penalty has to be added when S1 and S2 are identical,
        //      even if they are not perfectly complementary.
        char *space;
        int self_comp = 0;
        //char rev_comp_of_first_part[MAXSLEN/2];        
        space = strstr (sequence, " ");
        if (space != NULL)
        {
            int space_position;
            space_position = space - sequence;
            char first_part[MAXSLEN/2];
            strcpy (first_part, sequence);
            first_part[space_position] = '\0';
            //reverse_complement_of_seq (first_part, rev_comp_of_first_part);
            // Mirela comment added on Dec 14, 2007: the self-complementary sequences can have mismatches, so the following was incorrect
            // if (strcmp (first_part, space+1) == 0 && strstr (structure, ".") == NULL)
            if (strcmp (first_part, space+1) == 0)  // && strstr (structure, ".") == NULL)
            // they are identical and are not perfectly complementary
            //if (strcmp (first_part, space+1) == 0 && strcmp (first_part, rev_comp_of_first_part) == 0)
            {
                //printf ("self-comp: %s\t%s\n", sequence, structure);
                dg_37 -= 0.43;
                self_comp = 1;
            }    
        }
        
        // compute the lower bound and upper bound
        if (err_dg_37 < ABERANT && std_dg_37 < ABERANT)
        {
            printf ("One of err_dg_37 and std_dg_37 hs to be ABERANT. err_dg_37 = %.2lf, std_dg_37 = %.2lf\n", err_dg_37, std_dg_37);
            exit (1);
        }
        if (err_dg_37 < ABERANT)
        {
            //margin = dg_37 * err_dg_37/100;    // it gets infeasible
            margin = dg_37 * err_dg_37/100;
            if (margin < 0) margin *= -1;
            std_dg_37 = fabs(dg_37) * err_dg_37/100;
        }
        else
        {
            // I assume a normal distribution
            // Thus 99.73% of the area is within three standard deviations.
            //margin = std_dg_37*3; // this gets unfeasible, try with 4
            margin = std_dg_37*3;
        }
        lb = dg_37 - margin;
        ub = dg_37 + margin;
      
        // Added on Mar 20, 2008: the minimum error is 0.1
        if (std_dg_37 < 0.1)   std_dg_37 = 0.1;
        if (use_published_errors)
            precision = 1/(2*std_dg_37*std_dg_37);
        else
            precision = 1;
                      
        //if (self_comp) printf ("SELF-COMP  ");
        //else           printf ("NON-SE-CO  "); 
        //printf ("%s %s %lf %lf\n", sequence, structure, dg_37, std_dg_37 < ABERANT? std_dg_37 : err_dg_37);
        if (strcmp (sequence0, "") != 0)
        {            
            count_each_structure_type (sequence0, structure0, "", counter0, f0, 1);            
        }
        // added on Apr 17, 2008
        else
        {
            for (i=0; i < num_params; i++)  counter0[i]=0;
            f0 = 0;            
        }
        
        // now I have the real sequence and structure
        // store the real structure in matrix_A
        count_each_structure_type (sequence, structure, "", counter_min, f_min, 1);

        // do correction for Giese_Serra_1998 structures: the dangling ends should actually be tstackh, otherwise it's not full rank

        // THIS SHOULD NOT BE THERE (no issue with full rank or not, we have regularizer).
        // IF THE MODEL DOESN'T HAVE DANGLING ENDS, IT NEVER GETS IN HERE, AND IF IT DOES, IT GETS IN HERE.
        //  THIS COULD CAUSE THE c vector of the THERMO TERM TO BE DIFFERENT!!
        // REMOVED ON Sep 2, 2008.
        /*
        if ((strstr (exp_id, "Giese_Serra_1998") != NULL) ||
            (strstr (exp_id, "Dale_Serra_2000") != NULL) ||
            (strstr (exp_id, "Serra_Turner_1994") != NULL) ||
            (strstr (exp_id, "Vecenie_Serra_2004") != NULL) ||
            (strstr (exp_id, "Vecenie_Serra_2006") != NULL) )
        {         
            int bi,bj,bk,bl;   
            char *b;
            char stype[50];
            int index;
            int to_be_adjusted = 0;
            for (i=0; i < num_params; i++)
            {
                if ( (b = strstr (string_params[i], "dangle")) != NULL && counter_min[i] > 0)
                {
                    to_be_adjusted = 1;
                    counter_min[i] = 0;
                    // has dangle
                    //printf ("*** %s ", string_params[i]);
                    if (b[7] == 't')    // it's top
                    {
                        if (b[11] == '0')         bi=0;
                        else if (b[11] == '1')    bi=1;
                        else if (b[11] == '2')    bi=2;
                        else if (b[11] == '3')    bi=3;
                        if (b[14] == '0')         bj=0;
                        else if (b[14] == '1')    bj=1;
                        else if (b[14] == '2')    bj=2;
                        else if (b[14] == '3')    bj=3;
                        if (b[17] == '0')         bk=0;
                        else if (b[17] == '1')    bk=1;
                        else if (b[17] == '2')    bk=2;
                        else if (b[17] == '3')    bk=3;
                                                
                        //printf ("top i=%d j=%d k=%d ", bi, bj, bk);
                    }
                    else
                    {
                        if (b[17] == '0')         bl=0;
                        else if (b[17] == '1')    bl=1;
                        else if (b[17] == '2')    bl=2;
                        else if (b[17] == '3')    bl=3;                        
                        //printf ("bot ");
                    }
                    //break;                    
                } 
            }
            if (to_be_adjusted)
            {
                sprintf (stype, "tstackh[%d][%d][%d][%d]", bi, bj, bk, bl);
                //printf (" %s ", stype);
                index = structure_type_index (stype);
                counter_min [index] = 2;
                //printf (" %d %s %s\n", index, sequence, structure);
            }                
        }
        */
        
        char lb_constraint[5000];  
              
        if (strcmp (solver, "CPLEX") ==0)
        {
//printf ("==================\nEXP ID = %s\n", exp_id);
        
            int l=0;
            // first write the lower bound
            //sprintf (lb_constraint, "    tc_%s: ", exp_id);    // thermodynamic constraints
            // get the id only, it's too long otherwise
            char idid[10];
            for (int ii=0; ii < strlen (exp_id); ii++)
            {
                if (exp_id[ii] == '_')
                {
                    idid[ii] = '\0';
                    break;
                }
                idid[ii] = exp_id[ii];
            }

            for (i=0; i < num_params; i++)
            {
// if (counter_min[i] != 0)
// {
// int augi = i;
// if (num_params == 315 && i >= 260)  augi += 48;
// printf ("count_min[%d] = %.2lf, string_params = %s\n", augi, counter_min[i], string_params[i]);
// }                       
                if (counter_min[i] - counter0[i] != 0)
                {

                    param_met[i] = 1;

                    if (compute_second_order_matrix)
                    {
                        for (j=0; j < num_params; j++)
                        {
                            if (counter_min[j]-counter0[j] != 0)
                            {
                                second_order_matrix[i][j] += precision * (counter_min[i]-counter0[i])*(counter_min[j]-counter0[j]);
                            }
                        }
                        first_order[i] += -2*precision*(counter_min[i]-counter0[i])*(dg_37-(f_min-f0));
                        zero_order += precision*(dg_37-(f_min-f0))*(dg_37-(f_min-f0));
                    }
                }    
            }    
            
        }    // end if CPLEX
        //fclose (problem);
        //exit(1);
    }
    
    // TODO: in the file write_optimization_for_regression from ../RNA-thermodynamic-set, we make some of the variable equal to Turner values, so that the matrix becomes full rank. I don't think we need to do this here.
    

    fclose (xml);

    PARAMTYPE params_t99[MAXNUMPARAMS];
    save_parameters_in_array(params_t99);   


    if (compute_second_order_matrix)
    {
        FILE *matrix_file;
        if ((matrix_file = fopen (second_order_filename, "w")) == NULL)
        {
            printf ("Cannot open file %s\n", second_order_filename);
            exit (0);
        }
        for (i=0; i < num_params; i++)
        {
            //if (param_met[i])
            {
                for (j=0; j < num_params; j++)
                {
                    //if (param_met[j])
                        fprintf (matrix_file, "%.2lf ", second_order_matrix[i][j]);
                }
                fprintf (matrix_file, "\n");
            }
        }
        fclose (matrix_file);
        for (i=0; i < num_params; i++)
            delete [] second_order_matrix[i];
        delete [] second_order_matrix;

        FILE *first_order_file;
        if ((first_order_file = fopen (first_order_filename, "w")) == NULL)
        {
            printf ("Cannot open file %s\n", first_order_filename);
            exit (0);
        }
        for (i=0; i < num_params; i++)
        {
            fprintf (first_order_file, "%.15lf\n", first_order[i]);
        }
        // PRINT the zero_order at the end of first_order_file
        fprintf (first_order_file, "%.15lf\n", zero_order);
        fclose (first_order_file);
        
        delete [] first_order;
    }

    // print the number of features that appear in this thermo set
    int num_features = 0;
    for (i=0; i < num_params; i++)
        if (param_met[i])   num_features++;
    printf ("Total number of features in this XML set is %d\n", num_features);

    return 0;
}
